from __future__ import annotations

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from sklearn.base import BaseEstimator, OneToOneFeatureMixin, TransformerMixin
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer, StandardScaler
from xarray import DataArray
from discontinuum.pipeline import MetadataManager



class LogPropagation(OneToOneFeatureMixin, TransformerMixin, BaseEstimator):
    """Transformer for proper uncertainty propagation of log transforms
    
    Eq: f = ln(A) -> sigma_f = sigma_A / A
    """

    def fit(self, A):
        """Store data from A for later division in the transform.

        Parameters
        ----------
        A : ndarray
        """
        self.A_ = A

        return self

    def transform(self, X):
        """
        Propagate the uncertainty.

        Parameters
        ----------
        X : ndarray

        Returns
        -------
        ndarray
        """
        return X / self.A_


class LogUncertaintyPipeline(Pipeline):
    """Pipeline to transform (propagate) uncertainty to log space"""

    def __init__(self):
        super().__init__(
            steps=[
                ("metadata", MetadataManager()),
                ("log", LogPropagation()),
                ("scaler", StandardScaler(with_mean=False)),
                (
                    "square",
                    FunctionTransformer(
                        func=np.square,
                        inverse_func=np.sqrt,
                        check_inverse=False,
                    ),
                ),
                (
                    "abs",
                    FunctionTransformer(
                        func=np.abs,
                        inverse_func=np.abs,
                        check_inverse=False,
                    ),
                ),
            ]
        )



import numpy as np
from discontinuum.pipeline import TimeTransformer


def test_time_transform():
    """ Test the TimeTransformer class.
    """
    # Create sample data
    data = np.array(
        ["2022-01-01", "2022-02-01", "2022-03-01"],
        dtype="datetime64[ns]"  # ns precision
        )
    expected_result = np.array([2022. , 2022.08493151, 2022.16164384])

    # Create TimeTransformer instance
    transformer = TimeTransformer()

    # Perform transform
    transformed_data = transformer.transform(data)

    # Perform inverse transform
    inverse_transformed_data = transformer.inverse_transform(transformed_data)

    # Check if the transformed data matches the expected result
    np.testing.assert_array_almost_equal(transformed_data, expected_result, decimal=6)

    # Check if the inverse transformed data matches the original data
    np.testing.assert_equal(data, inverse_transformed_data)



"""Data transformations to improve optimization"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pymc as pm
from xarray import DataArray

from discontinuum.engines.base import BaseModel, is_fitted

if TYPE_CHECKING:
    from typing import Dict, Optional

    from xarray import Dataset


class LatentPyMC(BaseModel):
    def __init__(
        self,
        model_config: Optional[Dict] = None,
    ):
        """ """
        pass

    def fit(self, covariates, target=None):
        pass


class MarginalPyMC(BaseModel):
    def __init__(
        self,
        model_config: Optional[Dict] = None,
    ):
        """ """
        pass

    def fit(self, covariates: Dataset, target: Dataset, method: str = "BFGS"):
        """Fit the model to data.

        Parameters
        ----------
        covariates : Dataset
            Covariates for prediction.
        target : Dataset
            Target data for prediction.
        method : str, optional
            Optimization method. The default is "BFGS".
        """
        self.is_fitted = True
        # set up the data manager (self.dm)
        self.dm.fit(target=target, covariates=covariates)
        self.X = self.dm.X
        self.y = self.dm.y

        self.model = self.build_model(self.X, self.y)

        self.mp = pm.find_MAP(method=method, model=self.model)

    @is_fitted
    def predict(self, covariates, diag=True, pred_noise=False) -> DataArray:
        """Uses the fitted model to make predictions on new data."""
        Xnew = self.dm.Xnew(covariates)

        mu, var = self.gp.predict(
            Xnew,
            point=self.mp,
            diag=diag,
            pred_noise=pred_noise,
            model=self.model,
        )

        target = self.dm.y_t(mu)
        # TODO the reshape should be done in the pipeline
        se = self.dm.error_pipeline.inverse_transform(var.reshape(-1, 1))

        return target, se

    @is_fitted
    def predict_grid(self, covariate: str, index="time", t_step=12):
        """Predict on a grid of points.

        Parameters
        ----------
        covariate_dim : int, optional
            Dimension to predict on. The default is 1.
        t_step : int, optional
            Time steps per year. The default is 12.
        """
        time_dim = self.dm.get_dim(index)
        covariate_dim = self.dm.get_dim(covariate)

        x_max = self.dm.X.max(axis=0)
        x_min = self.dm.X.min(axis=0)
        x_range = x_max - x_min

        n_cov = 18
        n_time = np.round(x_range[time_dim] * t_step).astype(int)

        x_time = np.linspace(x_min[time_dim], x_max[time_dim], n_time)
        x_cov = np.linspace(x_min[covariate_dim], x_max[covariate_dim], n_cov)

        # TODO check dependency
        # tested with this on WSL with pymc v5.14.0
        # X_grid = pm.math.cartesian(x_cov[:, None], x_time[None, :])
        X_grid = pm.math.cartesian(x_time, x_cov)

        mu, _ = self.gp.predict(
            X_grid,
            point=self.mp,
            diag=True,
            pred_noise=True,
            model=self.model,
            )

        target = self.dm.y_t(mu)
        # TODO return a Dataset with the correct shape
        target = target.data.reshape(n_time, n_cov)
        t_pipe = self.dm.covariate_pipelines["time"]
        index = t_pipe.inverse_transform(x_time.reshape(-1, 1))

        c_pipe = self.dm.covariate_pipelines[covariate]
        covariate = c_pipe.inverse_transform(x_cov.reshape(-1, 1))

        return target, index, covariate

    @is_fitted
    def sample(self,
               covariates,
               n=1000,
               diag=False,
               pred_noise=False,
               method="cholesky",
               tol=1e-6,
               ) -> DataArray:
        """Sample from the posterior distribution of the model.

        Parameters
        ----------
        covariates : Dataset
            Covariates for prediction.
        n : int, optional
            Number of samples to draw.
        method{ ‘svd’, ‘eigh’, ‘cholesky’}, optional
            Method to use for covariance matrix decomposition. The default is
            ‘cholesky’.
        tol : float, optional
            Tolerance when checking the singular values in covariance matrix.
            The default is 1e-6.
        """
        Xnew = self.dm.Xnew(covariates)
        mu, cov = self.gp.predict(
            Xnew,
            point=self.mp,
            diag=diag,
            pred_noise=pred_noise,
            model=self.model,
        )

        rng = np.random.default_rng()
        sim = rng.multivariate_normal(mu, cov, size=n, method=method, tol=tol)

        # TODO modify transform to handle samples/draws HACK
        temp = self.dm.y_t(sim)
        data = temp.data.reshape(n, -1)
        attrs = temp.attrs
        da = DataArray(
            data,
            coords={"time": covariates.time, "draw": np.arange(n)},
            dims=["draw", "time"],
            attrs=attrs,
        )

        return da

    def build_model(self, X, y, **kwargs):
        """
        TODO: move this to parent?

        Creates an instance of pm.Model based on provided data and
        model_config, and attaches it to self.

        The subclass method must instantiate self.model and self.gp.

        Raises
        ------
        NotImplementedError
        """
        self.model = None
        self.gp = None

        raise NotImplementedError(
            "This method must be implemented in a subclass"
            )



"""Plotting functions"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from discontinuum.engines.base import is_fitted
from discontinuum.plot import BasePlotMixin
from xarray.plot.utils import label_from_attrs

if TYPE_CHECKING:
    from typing import Dict, Optional

    from matplotlib.pyplot import Axes
    from xarray import Dataset


DEFAULT_FIGSIZE = (5, 5)
NARROW_LINE = 1
REGULAR_LINE = NARROW_LINE * 1.5


class LoadestPlotMixin(BasePlotMixin):
    """Mixin plotting functions for Model class"""
    @is_fitted
    def plot_flux(self, covariates: Dataset, ax: Optional[Axes] = None):
        """Plot predicted flux versus time.

        Parameters
        ----------
        covariates : Dataset
            Covariates.

        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.
        """
        # TODO FINISH
        ax = self.setup_plot(ax)
        mu, se = self.predict(covariates, diag=True, pred_noise=True)

        flux = self.daily_flux(mu, covariates["flow"])
        flux_se = self.daily_flux(se, covariates["flow"])

        ci = 1.96 * flux_se

        flux.plot.line(ax=ax, lw=1, zorder=2)

        ax.fill_between(
            flux["time"], (flux - ci), (flux + ci), color="b", alpha=0.1, zorder=1
        )

        return ax

    @is_fitted
    def contourf(
        self,
        covariate: str = "flow",
        ax: Optional[Axes] = None,
        cbar_kwargs: Optional[Dict] = None,
        y_scale: str = "log",
        **kwargs,
    ):
        """Plot contourf

        TODO: plot any pair of variables on x and y axes.

        Parameters
        ----------
        covariate : str
            Covariate to plot.
        ax : Axes, optional
            Pre-defined matplotlib axes.
        cbar_kwargs : dict, optional
            Colorbar keyword arguments. The default is None.
        y_scale : str, optional
            y-axis scale. The default is "log".
        kwargs : dict
            Contourf keyword arguments.


        """
        ax = self.setup_plot(ax)

        if cbar_kwargs is None:
            cbar_kwargs = {}

        cbar_defaults = {"rotation": 270, "labelpad": 25}
        cbar_defaults.update(cbar_kwargs)

        ax = self.setup_plot(ax)
        ax.set_yscale(y_scale)

        target, time, cov = self.predict_grid(covariate=covariate, t_step=12)
        X2, X1 = np.meshgrid(cov, time)  # might need to flip these
        cs = ax.contourf(X1, X2, target, **kwargs)

        ax.set_ylabel(label_from_attrs(self.dm.data.covariates[covariate]))
        ax.set_xlabel("Year")

        fig = ax.get_figure()
        cbar = fig.colorbar(cs, ax=ax)
        cbar.ax.set_ylabel(
            label_from_attrs(self.dm.data.target),
            **cbar_defaults,
        )
        return ax

    @is_fitted
    def countourf_diff(self, ax: Optional[Axes] = None):
        pass



"""Helper functions for pulling USGS data."""

from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import TYPE_CHECKING

import pandas as pd
import xarray as xr
from dataretrieval import nwis, wqp
from discontinuum.providers.base import MetaData

if TYPE_CHECKING:
    # from pandas import DataFrame
    from typing import Dict, List, Optional, Union

    from xarray import Dataset
    from pandas import DataFrame

CFS_TO_M3 = 0.0283168


@dataclass
class USGSParameter:
    pcode: str
    standard_name: str
    long_name: Optional[str] = None
    units: Optional[str] = None
    suffix: Optional[str] = None
    conversion: Optional[float] = 1.0

    @property
    def ppcode(self):
        """
        Return the parameter code with a 'p' prefix, which is used by the QWData
        service.

        """
        return "p" + self.pcode

    @property
    def name(self):
        """
        Alias for standard_name.
        """
        return self.standard_name


USGSFlow = USGSParameter(
    pcode="00060",
    standard_name="flow",
    long_name="Streamflow",
    units="cubic meters per second",
    suffix="_Mean",
    conversion=0.0283168,
)


def get_parameters(
        pcodes: Dict[str, str],
        ) -> Union[USGSParameter, List[USGSParameter]]:
    """Get USGS parameters from a list of parameter codes.

    Parameters
    ----------
    pcodes : List[str]
        List of USGS parameter codes.

    Returns
    -------
    List[USGSParameter]
        List of USGS parameters.
    """
    lookup_name = {value: key for key, value in pcodes.items()}
    pcode_list = [v for _, v in pcodes.items()]

    df, _ = nwis.get_pmcodes(pcode_list)

    # convert to USGSParameter
    params = []
    for _, row in df.iterrows():
        params.append(
            USGSParameter(
                pcode=row["parameter_cd"],
                standard_name=lookup_name[row["parameter_cd"]],
                long_name=row["SRSName"],
                units=row["parm_unit"],
            )
        )

    if len(params) == 1:
        params = params[0]

    return params


def get_metadata(site: str) -> MetaData:
    """Get site metadata from the USGS NWIS API.

    Parameters
    ----------
    site : str
        USGS site number.

    Returns
    -------
    pandas.DataFrame
        Dataframe with the site information.
    """
    df, _ = nwis.get_info(sites=site)
    row = df.iloc[0]

    return MetaData(
        id=row["site_no"],
        name=row["station_nm"],
        latitude=row["dec_lat_va"],
        longitude=row["dec_long_va"],
    )


# TODO pass a dict not a list of params
def get_daily(
    site: str,
    start_date: str,
    end_date: str,
    params: Union[List[USGSParameter], USGSParameter] = USGSFlow,
    **kwargs,
) -> Dataset:
    """Get daily data from the USGS NWIS API.

    Parameters
    ----------
    site : str
        USGS site number.
    start_date : str
        Start date in the format 'yyyy-mm-dd'.
    end_date : str
        End date in the format 'yyyy-mm-dd'.
    params : List[USGSParameter], optional
        List of parameters to retrieve. The default is flow only `[USGSFlow]`.
    kwargs : Dict
        Additional keyword arguments to pass to the NWIS API.

    Returns
    -------
    Dataset
        Dataset with the requested data.
    """
    if not isinstance(params, list):
        params = [params]

    pcodes = [param.pcode for param in params]

    df, _ = nwis.get_dv(
        sites=site,
        start=start_date,
        end=end_date,
        parameterCd=pcodes,
        **kwargs,
        )

    return format_nwis_daily(df, site, params)


def format_wqp_date(date: str) -> str:
    """Reformat date from 'YYYY-MM-DD' to 'MM-DD-YYYY'."""
    return "-".join(date.split("-")[1:] + [date.split("-")[0]])


def format_nwis_daily(
        df: DataFrame,
        site_id: Optional[str] = None,
        params: Union[List[USGSParameter], USGSParameter] = [USGSFlow],
) -> Dataset:
    """
    Format results of dataretrieval.nwis.get_dv.

    Parameters
    ----------
    df : DataFrame
        Dataframe returned from dataretrieval.nwis.get_dv.
    site_id : str, optional
        USGS site number for populating metadata. The default is None.
    params : List[USGSParameter], optional
        List of parameters to retrieve. The default is flow only `[USGSFlow]`.
    """
    if not isinstance(params, list):
        params = [params]

    df = df.copy()
    # rename columns
    df = df.rename(
        columns={
            param.pcode + param.suffix: param.name for param in params}
    )
    # drop columns
    df = df[[param.name for param in params]]
    # remove timezone for xarray compatibility
    df.index = df.index.tz_localize(None)

    ds = xr.Dataset.from_dataframe(df)
    # rename "datetime" to "time", which is xarray convention
    ds = ds.rename({"datetime": "time"})
    # ds["date"] = ds["date"].dt.date

    # set metadata
    if site_id:
        ds.attrs = get_metadata(site_id).__dict__

    for param in params:
        ds[param.name] = ds[param.name] * param.conversion
        # xarray metadata assignment must come after all other operations
        ds[param.name].attrs = param.__dict__

    return ds


def format_wqp_samples(
        df: DataFrame,
        name: str = "concentration",
        pcode: Optional[str] = None,
) -> Dataset:
    """Format results of dataretrieval.wqp.get_results.

    Parameters
    ----------
    df : DataFrame
        Dataframe returned from dataretrieval.wqp.get_results.
    name : str
        Short name for the parameter. Default is 'concentration'.
    pcode : str
        USGS parameter code for populating metadata.

    Returns
    -------
    Dataset
    """
    # create datetime index
    df.index = pd.to_datetime(
        df["ActivityStartDate"] + " " + df["ActivityStartTime/Time"]
    )

    df[name] = df["ResultMeasureValue"].astype(float)
    df.index.name = "time"
    df.index = df.index.tz_localize(None)

    # create xarray dataset and remove unnecessary columns
    # TODO include censoring flag
    ds = xr.Dataset.from_dataframe(df[[name]])

    # drop censored values and warn user
    if any(ds[name].isnull()):
        ds = ds.dropna(dim="time")
        warnings.warn(
            "Censored values have been removed from the dataset.",
            stacklevel=1,
            )

    if pcode:
        attrs = get_parameters({name: pcode})
        ds[name].attrs = attrs.__dict__

    return ds


def get_samples(
        site: str,
        start_date: str,
        end_date: str,
        characteristic: str,
        fraction: str,
        provider: str = "NWIS",
        name: str = "concentration",
        filter_pcodes: Optional[List[str]] = None,
        **kwargs,
) -> Dataset:
    """Get sample data from the Water Quality Portal API.

    Parameters
    ----------
    site : str
        Water Quality Portal site id; e.g., 'USGS-12345678'.
    start_date : str
        Start date in the format 'YYYY-MM-DD'.
    end_date : str
        End date in the format 'YYYY-MM-DD'.
    characteristic : str
        The name of the parameter to retrieve.
    fraction : str
        The fraction of the parameter to retrieve. Options are 'Total',
        'Dissolved', 'Suspended'.
    provider : str
        The data provider. Options are 'NWIS' or 'STORET'.
    name : str
        Short name for the parameter. Default is 'concentration'.
    filter_pcodes : List[str]
        List of parameter codes to filter the dataset. 
    kwargs : Dict
        Additional keyword arguments to pass to the WQP API.
    """
    if fraction and fraction not in ["Total", "Dissolved", "Suspended"]:
        raise ValueError("Fraction must be 'Total', 'Dissolved', 'Suspended'")

    if provider not in ["NWIS", "STORET"]:
        raise ValueError("Provider must be 'NWIS' or 'STORET'")

    if provider == "NWIS" and not site.startswith("USGS-"):
        site = "USGS-" + site

    # reformat dates from 'YYYY-MM-DD' to 'MM-DD-YYYY'
    start_date = format_wqp_date(start_date)
    end_date = format_wqp_date(end_date)

    df, _ = wqp.get_results(
        siteid=site,
        startDateLo=start_date,
        startDateHi=end_date,
        characteristicName=characteristic,
        provider=provider,
        **kwargs,
        )

    # filter by fraction
    if fraction:
        df = df[df["ResultSampleFractionText"] == fraction]

    if provider == "NWIS":
        if filter_pcodes:
            # filter df by list of pcodes
            df = df[df["USGSPCode"].isin(filter_pcodes)]
            pcode = filter_pcodes[0]

        elif len(set(df["USGSPCode"])) != 1:
            raise ValueError("Multiple parameters returned from NWIS.")

        else:
            # only one pcode returned, so take the first entry
            pcode = df["USGSPCode"].iloc[0]

        pcode = str(pcode).zfill(5)

    else:
        pcode = None

    ds = format_wqp_samples(df, name, pcode)

    if provider == "NWIS":
        # strip the "USGS-" prefix from the site number
        site_id = site[5:]
        ds.attrs = get_metadata(site_id).__dict__

    if provider == "STORET":
        # TODO check for "USGS-" prefix, then get metadata
        pass

    return ds


def get_qwdata_samples(
    site: str,
    start_date: str,
    end_date: str,
    pcode: str,
    name: str = "concentration",
) -> Dataset:
    """Get sample data from the USGS NWIS API.

    WARNING: The QWData service is deprecated and no longer receives data.

    Parameters
    ----------
    site : str
        USGS site number.
    start_date : str
        Start date in the format 'yyyy-mm-dd'.
    end_date : str
        End date in the format 'yyyy-mm-dd'.
    pcode : str
        USGS parameter to retrieve.
    name : str
        Short name for the parameter.

    Returns
    -------
    pandas.DataFrame
        Dataframe with the requested sample data.
    """
    warnings.warn(
        "The QWData service is deprecated and no longer receives data.",
        DeprecationWarning,
    )

    df, _ = nwis.get_qwdata(
        sites=site,
        start=start_date,
        end=end_date,
        parameterCd=pcode,
    )
    attrs = get_parameters({name: pcode})
    ppcode = "p" + pcode

    # check if data are strings
    if df[ppcode].dtype == "O":
        # remove "<" and ">" from values and convert to float
        # TODO handle censoring
        df[ppcode] = df[ppcode].str.extract(r"(\d+.\d+)", expand=False).astype(float)

    df = df.rename(columns={ppcode: name})

    df = df[[name]]
    # remove timezone for xarray compatibility
    df.index = df.index.tz_localize(None)

    ds = xr.Dataset.from_dataframe(df)
    # rename "datetime" to "time", which is xarray convention
    ds = ds.rename({"datetime": "time"})

    ds.attrs = get_metadata(site).__dict__
    ds[name].attrs = attrs.__dict__

    return ds



import numpy as np
import pymc as pm
from discontinuum.engines.pymc import MarginalPyMC

from loadest_gp.models.base import LoadestDataMixin, ModelConfig
from loadest_gp.plot import LoadestPlotMixin


class LoadestGPMarginalPyMC(LoadestDataMixin, LoadestPlotMixin, MarginalPyMC):
    """
    Gaussian Process implementation of the LOAD ESTimation (LOADEST) model

    This model currrently uses the marginal likelihood implementation, which is
    fast but does not account for censored data. Censored data require a slower
    latent variable implementation.
    """
    def __init__(
            self,
            model_config: ModelConfig = ModelConfig(),
    ):
        """ """
        super(MarginalPyMC, self).__init__(model_config=model_config)
        self.build_datamanager(model_config)

    def build_model(self, X, y) -> pm.Model:
        """Build marginal likelihood version of LoadestGP
        """
        n_d = X.shape[1]  # number of dimensions
        dims = np.arange(n_d)
        time_dim = [dims[0]]
        cov_dims = dims[1:]

        with pm.Model() as model:
            # seasonal trend
            eta_per = pm.HalfNormal("eta_per", sigma=1, initval=1)
            ls_pdecay = pm.Gamma("ls_pdecay", alpha=10, beta=1)
            # https://peterroelants.github.io/posts/gaussian-process-kernels/
            period = pm.Normal("period", mu=1, sigma=0.05)
            ls_psmooth = pm.Gamma("ls_psmooth", alpha=4, beta=3)

            cov_seasonal = (
                eta_per**2
                * pm.gp.cov.Periodic(n_d, period, ls_psmooth, active_dims=time_dim)
                * pm.gp.cov.Matern52(n_d, ls_pdecay, active_dims=time_dim)
            )
            gp_seasonal = pm.gp.Marginal(cov_func=cov_seasonal)

            # longterm trend
            eta_trend = pm.Exponential(
                "eta_trend", scale=1.5
            )  # Exponential might dampen outlier effects
            ls_trend = pm.Gamma("ls_trend", alpha=4, beta=1)
            cov_trend = eta_trend**2 * pm.gp.cov.ExpQuad(n_d, ls_trend, active_dims=time_dim)
            gp_trend = pm.gp.Marginal(cov_func=cov_trend)

            # covariate trend
            # could include time with a different prior on ls
            eta_covariates = pm.HalfNormal("eta_covariates", sigma=2)
            ls_covariates = pm.Gamma(
                "ls_covariates",
                alpha=2,
                beta=3,
                initval=[0.5],
                shape=n_d-1,  # exclude time
                )
            cov_covariates = eta_covariates**2 \
                * pm.gp.cov.ExpQuad(n_d, ls=ls_covariates, active_dims=cov_dims)
            gp_covariates = pm.gp.Marginal(cov_func=cov_covariates)

            # residual trend
            eta_res = pm.Exponential("eta_res", scale=0.2)
            ls_res = pm.Gamma("ls_res", alpha=2, beta=10, shape=n_d)
            cov_res = eta_res**2 * pm.gp.cov.Matern32(n_d, ls_res, active_dims=dims)
            gp_res = pm.gp.Marginal(cov_func=cov_res)

            gp = gp_trend + gp_seasonal + gp_covariates + gp_res

            # noise_model
            # set to a small value to ensure cov is positive definite
            sigma = 0.1
            # or
            # sigma = pm.HalfNormal("sigma", sigma=0.05)
            cov_noise = pm.gp.cov.WhiteNoise(sigma=sigma)

            y_ = gp.marginal_likelihood("y", X=X, y=y, sigma=cov_noise)  # noqa: F841

            self.gp = gp

        return model



from __future__ import annotations

from typing import TYPE_CHECKING

from dataclasses import dataclass

from discontinuum.data_manager import DataManager
from discontinuum.pipeline import (
    LogStandardPipeline,
    LogErrorPipeline,
    NoOpPipeline,
    StandardPipeline,
    StandardErrorPipeline,
    TimePipeline,
)
# from rating_gp.pipeline import LogUncertaintyPipeline

if TYPE_CHECKING:
    from typing import Literal

@dataclass
class ModelConfig:
    """ """
    transform: Literal["log", "standard"] = "log"


class RatingDataMixin:
    """ """
    # TODO inheret a BaseModel class with abc methods like build_model
    def build_datamanager(
            self,
            model_config: ModelConfig = ModelConfig(),
            ):
        """ """
        covariate_pipelines = {
            "time": TimePipeline,
            "stage": NoOpPipeline
        }

        if model_config.transform == "log":
            target_pipeline = LogStandardPipeline
            error_pipeline = LogErrorPipeline
        elif model_config.transform == "standard":
            target_pipeline = StandardPipeline
            error_pipeline = StandardErrorPipeline
        else:
            raise ValueError(
                "Model config transform must be 'log' or 'standard'."
            )

        self.dm = DataManager(
            target_pipeline=target_pipeline,
            error_pipeline=error_pipeline,
            covariate_pipelines=covariate_pipelines
        )



"""Helper functions for pulling USGS data."""

from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import TYPE_CHECKING

import pandas as pd
import xarray as xr
from dataretrieval import nwis, wqp
from discontinuum.providers.base import MetaData
from loadest_gp.providers.usgs import get_metadata

if TYPE_CHECKING:
    # from pandas import DataFrame
    from typing import Dict, List, Optional, Union

    from xarray import Dataset

FT_TO_M = 0.3048
FT3_TO_M3 = 0.0283168

@dataclass
class NWISColumn:
    column_name: str
    standard_name: str
    long_name: [Optional[str]] = None
    units: [Optional[str]] = None
    conversion: float = 1.0

    @property
    def name(self):
        """
        Alias for standard_name.
        """
        return self.standard_name


@dataclass
class USGSParameter:
    pcode: str
    standard_name: str
    long_name: Optional[str] = None
    units: Optional[str] = None
    suffix: Optional[str] = None
    conversion: Optional[float] = 1.0

    @property
    def ppcode(self):
        """
        Return the parameter code with a 'p' prefix, which is used by the QWData service.

        """
        return "p" + self.pcode

    @property
    def name(self):
        """
        Alias for standard_name.
        """
        return self.standard_name


USGSStage = USGSParameter(
    pcode="00065",
    standard_name="stage",
    long_name="Stream stage",
    units="meters",
    suffix="_Mean",
    conversion=FT_TO_M,
)

NWISStage = NWISColumn(
    column_name="gage_height_va",
    standard_name="stage",
    long_name="Stream stage",
    units="meters",
    conversion=FT_TO_M,
)

NWISDischarge = NWISColumn(
    column_name="discharge_va",
    standard_name="discharge",
    long_name="Stream discharge",
    units="cubic meters per second",
    conversion=FT3_TO_M3,
)

NWISDischargeUnc = NWISColumn(
    column_name="measured_rating_diff",
    standard_name="discharge_unc",
    long_name=("Stream discharge uncertainty estimated from qualitative "
               "measurement rating codes"),
    units="cubic meters per second",
    conversion=FT3_TO_M3,
)


def get_daily_stage(
    site: str,
    start_date: str,
    end_date: str,
) -> Dataset:
    """Get daily data from the USGS NWIS API.

    Parameters
    ----------
    site : str
        USGS site number.
    start_date : str
        Start date in the format 'yyyy-mm-dd'.
    end_date : str
        End date in the format 'yyyy-mm-dd'.
    params : List[USGSParameter], optional
        List of parameters to retrieve. The default is flow only `[USGSFlow]`.

    Returns
    -------
    Dataset
        Dataset with the requested data.
    """
    param = USGSStage
    df, _ = nwis.get_dv(
        sites=site,
        start=start_date,
        end=end_date,
        parameterCd=param.pcode,
        )

    if len(df) == 0:
        raise ValueError("No daily stage data is available for USGS site "
                         f"number: {site}")

    # rename columns
    df = df.rename(columns={param.pcode + param.suffix: param.name})
    # drop columns
    df = df[[param.name]]
    # remove timezone for xarray compatibility
    df.index = df.index.tz_localize(None)

    ds = xr.Dataset.from_dataframe(df)
    # rename "datetime" to "time", which is xarray convention
    ds = ds.rename({"datetime": "time"})

    # set metadata
    ds.attrs = get_metadata(site).__dict__
    # convert units
    ds[param.name] = ds[param.name] * param.conversion
    # xarray metadata assignment must come after all other operations
    ds[param.name].attrs = param.__dict__

    return ds


def get_measurements(
        site: str,
        start_date: str,
        end_date: str,
):
    """Get discharge measurements from the USGS NWIS API.

    Parameters
    ----------
    site : str
        Water Quality Portal site id; e.g., 'USGS-12345678'.
    start_date : str
        Start date in the format 'YYYY-MM-DD'.
    end_date : str
        End date in the format 'YYYY-MM-DD'.
    """
    # df, _ = nwis.get_discharge_measurements(
    #     sites=site,
    #     start=start_date,
    #     end=end_date,
    #     format="rdb_expanded",
    # )
    # Need this till get_discharge_measurements update is uploaded
    response = nwis.query_waterdata(
        'measurements', ssl_check=True, format="rdb_expanded",
        site_no=site, begin_date=start_date, end_date=end_date,
    )
    df = nwis._read_rdb(response.text)

    # covert timezone to UTC? ignore for now
    df.index = pd.to_datetime(
        df["measurement_dt"],
        format="ISO8601",
    )
    df.index.name = "time"
    # df.index = df.index.tz_localize(None)
    df = df.rename(
        columns={
            NWISStage.column_name: NWISStage.standard_name,
            NWISDischarge.column_name: NWISDischarge.standard_name,
            }
        )
    # parse uncertainty from measured "measured_rating_diff"
    qualitycode_to_uncertainty_fraction = {
        'Excellent': '0.02',
        'Good': '0.05',
        'Fair': '0.08',
        'Poor': '0.12',
        'Unspecified': '0.12',
    }
    df['discharge_unc_frac'] = (df['measured_rating_diff']
                                .replace(qualitycode_to_uncertainty_fraction)
                                .astype(float))
    # set indirect measurements as 20% uncertain regardless of quality code
    df.loc[df['streamflow_method'] == 'QIDIR', 'discharge_unc_frac'] = 0.2
    # convert fractional uncertainty to uncertainty assuming the uncertainty
    # fraction is a 2 sigma gse interval. (GSE = frac + 1)
    # (GSE -> exp(sigma_ln(Q)))
    df['discharge_unc'] = df['discharge_unc_frac'] / 2 + 1

    # drop data that is <= 0 as we need all positive data
    df = df[(df['stage'] > 0) & (df['discharge'] > 0)]

    ds = xr.Dataset.from_dataframe(df[["stage", "discharge", "discharge_unc"]])

    for param in [NWISStage, NWISDischarge]:
        ds[param.name] = ds[param.name] * param.conversion
        ds[param.name].attrs = param.__dict__

    ds['discharge_unc'].attrs = NWISDischargeUnc.__dict__

    return ds



"""Plotting functions"""

from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from discontinuum.engines.base import is_fitted
from discontinuum.plot import BasePlotMixin
from scipy.stats import norm
import xarray as xr
from xarray import DataArray
from xarray.plot.utils import label_from_attrs

if TYPE_CHECKING:
    from typing import Dict, Optional

    from matplotlib.pyplot import Axes
    from xarray import Dataset


DEFAULT_FIGSIZE = (5, 5)


class RatingPlotMixin(BasePlotMixin):
    """Mixin plotting functions for Model class"""

    @is_fitted
    def plot_observed_rating(self, ax: Optional[Axes] = None, **kwargs):
        """Plot stage observations versus discharge observations.

        Parameters
        ----------
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.
        """
        ax = self.setup_plot(ax)

        data = xr.merge([self.dm.data.covariates["stage"],
                         self.dm.data.target])

        # This converts the input uncertainty from GSE to SE
        se_unc = np.log(self.dm.data.target_unc) * self.dm.data.target
        ax.errorbar(self.dm.data.covariates["stage"],
                    self.dm.data.target,
                    yerr=se_unc,
                    c="k",
                    marker='',
                    linestyle='',
                    capsize=2.5)
        data.plot.scatter(x='stage',
                          y='discharge',
                          hue="time",
                          add_colorbar=False,
                          add_legend=False,
                          marker='o',
                          edgecolors='face',
                          s=10,
                          ax=ax,
                          **kwargs)

        return ax

    @is_fitted
    def plot_rating(self,
                    covariates: Dataset,
                    ci: float = 0.95,
                    ax: Optional[Axes] = None):
        """Plot predicted discharge versus stage.

        Parameters
        ----------
        covariates : Dataset
            Covariates.
        ci : float, optional
            Confidence interval. The default is 0.95.
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.
        """
        ax = self.setup_plot(ax)

        mu, se = self.predict(covariates, diag=True, pred_noise=True)

        target = DataArray(
            mu,
            coords=[covariates.time],
            dims=["time"],
            attrs=self.dm.data.target.attrs,
        )

        # compute confidence bounds
        lower, upper = self.dm.error_pipeline.ci(target, se, ci=ci)

        ax.plot(covariates["stage"], target, lw=1, zorder=2)

        ax.fill_between(
            covariates['stage'],
            lower,
            upper,
            color="b",
            alpha=0.1,
            zorder=1,
        )

        self.plot_observed_rating(ax, zorder=3)

        return ax

    @is_fitted
    def plot_stage(self,
                   covariates: Dataset = None,
                   ax: Optional[Axes] = None,
                   **kwargs):
        """Plot observations versus time.

        Parameters
        ----------
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.

        """
        self.dm.data.covariates.plot.scatter(
            x='time',
            y='stage',
            hue='time',
            s=10,
            edgecolors="face",
            ax=ax,
            zorder=2,
            add_legend=False,
            add_colorbar=False,
            **kwargs,
        )

        if covariates is not None:
            covariates["stage"].plot.line(ax=ax, lw=1, zorder=1)

        return ax

    @is_fitted
    def plot_discharge(self,
                       covariates: Dataset = None,
                       ci: float = 0.95,
                       ax: Optional[Axes] = None
                      ):
        """Plot predicted discharge versus time.

        Parameters
        ----------
        covariates : Dataset, optional
            Covariates.
        ci : float, optional
            Confidence interval. The default is 0.95.
        ax : Axes, optional
            Pre-defined matplotlib axes.

        Returns
        -------
        ax : Axes
            Generated matplotlib axes.
        """
        ax = self.setup_plot(ax)

        if covariates is not None:
            mu, se = self.predict(covariates, diag=True, pred_noise=True)
    
            target = DataArray(
                mu,
                coords=[covariates.time],
                dims=["time"],
                attrs=self.dm.data.target.attrs,
            )
    
            # compute confidence bounds
            lower, upper = self.dm.error_pipeline.ci(target, se, ci=ci)

            target.plot.line(ax=ax, lw=1, zorder=2)

            ax.fill_between(
                target["time"],
                lower,
                upper,
                color="b",
                alpha=0.1,
                zorder=1,
            )

        self.dm.data.target.plot.scatter(
            x='time',
            y='discharge',
            hue='time',
            s=10,
            edgecolors="face",
            ax=ax,
            zorder=2,
            add_legend=False,
            add_colorbar=False,
        )

        return ax
